
/*
	HPT366 IDE module
	2000 Jens Kristian Jensen
	Based on genericpci.c by Be Inc. and Linux's hpt366.c by Andre Hedrick.
*/

#include "IDE.h"
#include <PCI.h>
#include <KernelExport.h>
#include <string.h>
#include <malloc.h>
#include "lendian_bitfield.h"
#include <ByteOrder.h>
#include <driver_settings.h>
#include <stdlib.h>

// Settings variables read from the driver settings file (~/config/settings/kernel/drivers/hpt366)

static int debug_errors = 1;		// Print out errors
static int debug_settings = 1;		// Print out settings
static int debug_finddevices = 1;	// Print out info when finding devices (looking for the controller among the PCI devices)
static int debug_interrupts = 0;	// Print out when waiting for interrupts and handling interrupts
static int debug_dma = 0;			// Print out when preparing, starting and finishing DMA
static int debug_prd = 0;			// Print out when creating, deleting or updating physical region descriptors
static int debug_verbose = 0;		// Print out when entering and leaving other functions
static int disable_dma = 0;			// If true, no DMA code is used and the driver reports that it cannot do DMA. PIO can still be used, though.
#define DMA_MODE_LENGTH 10
static char force_dma_mode[DMA_MODE_LENGTH];	// The user can force the driver to set the timing registers of the controller to a specific DMA mode.
									// The possible values are: DMA0, DMA1, DMA2, UDMA0, UDMA1, UDMA2, UDMA3, UDMA4 and AUTO (which uses the BIOS' settings). Case matters!

static pci_module_info	*pci;		// Pointer to pci module info (See docs for the PCI bus manager)

typedef struct
{
	uint32	address;		// bit 0 reserved
	uint16	byte_count;		// bit 0 reserved
	uint8	reserved_1;
	LBITFIELD8_2(
			reserved_2 : 7,
	        eot        : 1
	);
} prd_t;
const size_t 	prd_table_size = B_PAGE_SIZE;

/*
	The bus_info structure is used to describe a PCI IDE controller channel
	In the case of the HPT366, there are actually a whole PCI device for
	each bus_info, but that's another topic.
	When you need to access an IDE channel through its driver module, you
	ask for a "cookie" for that bus using get_nth_cookie. The cookie you get
	is actually a pointer to a bus_info structure and that pointer is passed
	as a parameter to subsequent calls to the driver's other functions.
*/

typedef struct bus_info_s
{
	struct bus_info_s *next;			// All bus_info structures are chained using a single linked lists.
	uint16	command_block_base_addr;
	uint16	control_block_register_addr;
	uint16	busmaster_base_addr;
	uint16	busmaster_status_addr;
	int		intnum;
	int		interrupt_gate;				// Used to see if the interrupt for a particular request has already been handled.
										// See comment in intwait.
	sem_id	mutex;
	bool	mutex_shared;
	sem_id	intsem;
	
	uint8	busmasterstatus;
	bool	shared_interrupt;
	uint32	bad_alignment_mask;

	area_id	prd_table_area;
	prd_t	*prd_table;
	uint32	prd_table_physical;

	bool dmainuse;
	bool dma_started;

	bool buslock;

	interrupt_handler handler;
	struct {
		uint8	bus;
		uint8	device;
		uint8	function;
	} pci_dev;
} bus_info;

static int buscount;
static bus_info *busses;

/* The following is heavily inspired by Linux               */
/* The data and code takes care of setting up the timing    */
/* registers of the HPT366 at three different PCI bus       */
/* clock speeds; 25, 33 (default) and 40MHz.                */
/* Converted to BeOS IDE driver API by Jens Kristian Jensen */

#	define XFER_UDMA_4	0x44	/* 0100|0100 */
#	define XFER_UDMA_3	0x43	/* 0100|0011 */
#	define XFER_UDMA_2	0x42	/* 0100|0010 */
#	define XFER_UDMA_1	0x41	/* 0100|0001 */
#	define XFER_UDMA_0	0x40	/* 0100|0000 */
#	define XFER_MW_DMA_2	0x22	/* 0010|0010 */
#	define XFER_MW_DMA_1	0x21	/* 0010|0001 */
#	define XFER_MW_DMA_0	0x20	/* 0010|0000 */
#	define XFER_SW_DMA_2	0x12	/* 0001|0010 */
#	define XFER_SW_DMA_1	0x11	/* 0001|0001 */
#	define XFER_SW_DMA_0	0x10	/* 0001|0000 */
#	define XFER_PIO_4	0x0C	/* 0000|1100 */
#	define XFER_PIO_3	0x0B	/* 0000|1011 */
#	define XFER_PIO_2	0x0A	/* 0000|1010 */
#	define XFER_PIO_1	0x09	/* 0000|1001 */
#	define XFER_PIO_0	0x08	/* 0000|1000 */
#	define XFER_PIO_SLOW	0x00	/* 0000|0000 */

struct chipset_bus_clock_list_entry {
	uint8		xfer_speed;
	unsigned int	chipset_settings;
};

struct chipset_bus_clock_list_entry forty_base [] = {

	{	XFER_UDMA_4	,	0x900fd943	},
	{	XFER_UDMA_3	,	0x900ad943	},
	{	XFER_UDMA_2	,	0x900bd943	},
	{	XFER_UDMA_1	,	0x9008d943	},
	{	XFER_UDMA_0	,	0x9008d943	},

	{	XFER_MW_DMA_2	,	0xa008d943	},
	{	XFER_MW_DMA_1	,	0xa010d955	},
	{	XFER_MW_DMA_0	,	0xa010d9fc	},

	{	XFER_PIO_4	,	0xc008d963	},
	{	XFER_PIO_3	,	0xc010d974	},
	{	XFER_PIO_2	,	0xc010d997	},
	{	XFER_PIO_1	,	0xc010d9c7	},
	{	XFER_PIO_0	,	0xc018d9d9	},
	{	0		,	0x0120d9d9	}
};

struct chipset_bus_clock_list_entry thirty_three_base [] = {

	{	XFER_UDMA_4	,	0x90c9a731	},
	{	XFER_UDMA_3	,	0x90cfa731	},
	{	XFER_UDMA_2	,	0x90caa731	},
	{	XFER_UDMA_1	,	0x90cba731	},
	{	XFER_UDMA_0	,	0x90c8a731	},

	{	XFER_MW_DMA_2	,	0xa0c8a731	},
	{	XFER_MW_DMA_1	,	0xa0c8a732	},	/* 0xa0c8a733 */
	{	XFER_MW_DMA_0	,	0xa0c8a797	},

	{	XFER_PIO_4	,	0xc0c8a731	},
	{	XFER_PIO_3	,	0xc0c8a742	},
	{	XFER_PIO_2	,	0xc0d0a753	},
	{	XFER_PIO_1	,	0xc0d0a7a3	},	/* 0xc0d0a793 */
	{	XFER_PIO_0	,	0xc0d0a7aa	},	/* 0xc0d0a7a7 */
	{	0		,	0x0120a7a7	}
};

struct chipset_bus_clock_list_entry twenty_five_base [] = {

	{	XFER_UDMA_4	,	0x90c98521	},
	{	XFER_UDMA_3	,	0x90cf8521	},
	{	XFER_UDMA_2	,	0x90cf8521	},
	{	XFER_UDMA_1	,	0x90cb8521	},
	{	XFER_UDMA_0	,	0x90cb8521	},

	{	XFER_MW_DMA_2	,	0xa0ca8521	},
	{	XFER_MW_DMA_1	,	0xa0ca8532	},
	{	XFER_MW_DMA_0	,	0xa0ca8575	},

	{	XFER_PIO_4	,	0xc0ca8521	},
	{	XFER_PIO_3	,	0xc0ca8532	},
	{	XFER_PIO_2	,	0xc0ca8542	},
	{	XFER_PIO_1	,	0xc0d08572	},
	{	XFER_PIO_0	,	0xc0d08585	},
	{	0		,	0x01208585	}
};

/*
	Converts from PCI clock and DMA mode to hpt366 chipset timing
	Called from hpt366_tune_chipset
*/
static unsigned int pci_bus_clock_list (uint8 speed, struct chipset_bus_clock_list_entry * chipset_table)
{
	for ( ; chipset_table->xfer_speed ; chipset_table++)
		if (chipset_table->xfer_speed == speed)
			return chipset_table->chipset_settings;
	return 0x01208585;	// XFER_PIO_SLOW for 25MHz PCI bus
}

/*
	Given a PCI clock speed, return the DMA mode for a given timimg register setting
*/
static uint8 pci_dma_mode_list (unsigned int chipset_settings, struct chipset_bus_clock_list_entry * chipset_table)
{
	for ( ; chipset_table->xfer_speed ; chipset_table++)
		if (chipset_table->chipset_settings == chipset_settings)
			return chipset_table->xfer_speed;
	return 0xFF;	// None found
}

static int hpt366_tune_chipset (bus_info *bi, uint32 drive, uint8 speed)
{
// Master timing register address = 0x40, Slave = 0x44
	uint8 regtime		=  drive ? 0x44 : 0x40; // 0=Master, >0=Slave
	int			err;
	uint32 reg2;
	uint32 TimingRegister;
	uint8 busclock;
	int i;

	TimingRegister =
		pci->read_pci_config(bi->pci_dev.bus, bi->pci_dev.device,
		                     bi->pci_dev.function, regtime,
		                     4);

// Peek PCI clock speed from timing register (initially set by HPT366 BIOS)
// Remember that this is little endian, so offset 1 is the second last byte
// when the timing register is viewed as 32bit big endian
	busclock = 
		pci->read_pci_config(bi->pci_dev.bus, bi->pci_dev.device,
		                     bi->pci_dev.function, regtime+1,
		                     1);

	dprintf("HPT366_tune_chipset: Busclock is: 0x%X (for drive %d on bus 0x%x)\n",busclock,drive,bi);
	dprintf("HPT366_tune_chipset: Timing register is: 0x%X (for drive %d on bus 0x%x)\n",TimingRegister,drive,bi);

	switch(busclock) { // PCI bus clock
		case 0xd9: // 40MHz
			reg2 = pci_bus_clock_list(speed, forty_base);
			break;
		case 0x85: // 25MHz
			reg2 = pci_bus_clock_list(speed, twenty_five_base);
			break;
		case 0xa7: // 33MHz
		default:   // other
			reg2 = pci_bus_clock_list(speed, thirty_three_base);
			break;
	}
	
	if (false) // Condition should check if drive supports dword io
		reg2 |= 0x80000000;		// Set MSB (PIO buffer enable)
	else
		reg2 &= ~0x80000000;	// Clear MSB (PIO buffer enable)

	dprintf("reg2 is: 0X%X\n",reg2);


	pci->write_pci_config(bi->pci_dev.bus, bi->pci_dev.device,
	                     bi->pci_dev.function, regtime,
	                     4,reg2);

	// Linux configs drive for this speed too!
	// How is this done in BeOS?
	//err = ide_config_drive_speed(drive, speed);
}
/* End of Linux inspiration */

static bus_info *
get_nth_cookie(uint32 bus)
{
	int i;
	bus_info *cookie = busses;
	
	for(i=0; i<bus; i++) {
		if(cookie == NULL) {
			if(debug_errors)
				dprintf("IDE hpt366 -- get_nth_cookie: no cookie for bus %ld\n", bus);
			return NULL;
		}
		cookie = cookie->next;
	}
	if(debug_verbose)
		dprintf("IDE hpt366 -- get_nth_cookie: cookie for bus %d is 0x%08x\n", bus, cookie);
	return cookie;
}

static uint32
get_bus_count()
{
	if(debug_verbose)
		dprintf("IDE hpt366 -- get_bus_count returns %d\n",buscount);
	return buscount;
}

static int32
get_abs_bus_num(bus_info *bi)
{
	if(debug_verbose)
		dprintf("IDE hpt366 -- get_abs_bus_num returns -1\n");
	return -1;
}

static status_t
acquire_bus(bus_info *bi)
{
	status_t err;
	err = acquire_sem(bi->mutex);
	if(err == B_NO_ERROR)
		bi->buslock = true;
	if(debug_verbose)
		dprintf("IDE hpt366 -- acquire_bus returns %d (%s)\n",err, strerror(err));
	return err;
}

static status_t
release_bus(bus_info *bi)
{
	status_t err;
	bi->buslock = false;
	err=release_sem_etc(bi->mutex, 1, B_DO_NOT_RESCHEDULE);
	if(debug_verbose)
		dprintf("IDE hpt366 -- release_bus returns %d (%s)\n",err, strerror(err));
	return;
}

static status_t
write_command_block_regs(bus_info *bi, ide_task_file *tf, ide_reg_mask mask)
{
	int i;
	uint16 ioaddr = bi->command_block_base_addr;

	if(!bi->buslock) {
		if(debug_errors)
			dprintf("IDE hpt366 -- write_command_block_regs: bus not locked\n");
		return B_ERROR;
	}

	if((mask & ide_mask_command) && bi->dmainuse && bi->dma_started) {
		if(debug_errors)
			dprintf("IDE hpt366 -- write_command_block_regs: "
			        "error: only one command allowed\n"
			        "before calling finish_dma\n");
		return B_NOT_ALLOWED;
	}
	
	for(i=0; i<7; i++) {
		if((1<<i) & mask) {
			if(debug_verbose)
				dprintf("IDE hpt366 -- write_command_block_regs: "
						"write r[%d] = 0x%02x\n", i, tf->raw.r[i]);
			pci->write_io_8(ioaddr+1+i, tf->raw.r[i]);
		}
	}

	return B_NO_ERROR;
}

static status_t
read_command_block_regs(bus_info *bi, ide_task_file *tf, ide_reg_mask mask)
{
	int i;
	uint16 ioaddr = bi->command_block_base_addr;

	if(!bi->buslock) {
		if(debug_errors)
			dprintf("IDE hpt366 -- read_command_block_regs: bus not locked\n");
		return B_ERROR;
	}

	for(i=0; i<7; i++) {
		if((1<<i) & mask) {
			tf->raw.r[i] = pci->read_io_8(ioaddr+1+i);
			if(debug_verbose)
				dprintf("IDE hpt366 -- read_command_block_regs: "
			        "read r[%d] = 0x%02x\n", i, tf->raw.r[i]);
		}
	}
	return B_NO_ERROR;
}

static uint8
get_altstatus(bus_info *bi)
{
	uint8 val;
	uint16 altstatusaddr = bi->control_block_register_addr;
	if(!bi->buslock) {
		if(debug_errors)
			dprintf("IDE hpt366 -- get_altstatus: bus not locked\n");
	}
	val= pci->read_io_8(altstatusaddr);
	if(debug_verbose)
		dprintf("IDE hpt366 -- get_altstatus returns 0x%02x from bus 0x%x\n",val, bi);
	return val;
}

static void
write_device_control(bus_info *bi, uint8 val)
{
	uint16 device_control_addr = bi->control_block_register_addr;
	if(!bi->buslock) {
		if(debug_errors)
			dprintf("IDE hpt366 -- write_device_control: bus not locked\n");
		return;
	}
	if(debug_verbose)
		dprintf("IDE hpt366 -- write_device_control writes 0x%02X to bus 0x%x",val,bi);
	pci->write_io_8(device_control_addr, val);
}

static void
write_pio_16(bus_info *bi, uint16 *data, uint16 count)
{
	uint16 ioaddr = bi->command_block_base_addr;
	uint32 *dp = (uint32 *)data;
	uint32 *ep = (uint32 *)(data+count);
	
	if(!bi->buslock) {
		if(debug_errors)
			dprintf("IDE hpt366 -- write_pio_16: bus not locked\n");
		return;
	}
	if(debug_verbose)
		dprintf("IDE hpt366 -- write_pio_16: writing %d words\n",count);
	while(dp < ep) {
		pci->write_io_32(ioaddr, *(dp++));
	}
}

static void
read_pio_16(bus_info *bi, uint16 *data, uint16 count)
{
	uint16 ioaddr = bi->command_block_base_addr;
	uint32 *dp = (uint32 *)data;
	uint32 *ep = (uint32 *)(data+count);
	
	if(!bi->buslock) {
		dprintf("IDE hpt366 -- read_pio_16: bus not locked\n");
		return;
	}
	if(debug_verbose)
		dprintf("IDE hpt366 -- read_pio_16: reading %d words\n",count);
	while(dp < ep) {
		*(dp++) = pci->read_io_32(ioaddr);
	}
}

static status_t
intwait(bus_info *bi, bigtime_t timeout)
{
	status_t err;
	int32 count = -99;
	if(debug_interrupts)
		dprintf("IDE hpt366 -- intwait\n");
	get_sem_count(bi->intsem, &count);
	if(debug_interrupts)
		dprintf("IDE hpt366 -- intwait: semcount = %d, intgate = %d, timeout = 0x%X\n",
		        count, bi->interrupt_gate, timeout);
	if(!bi->buslock) {
		if(debug_errors)
			dprintf("IDE hpt366 -- intwait: bus not locked\n");
		return B_ERROR;
	}
	err = acquire_sem_etc(bi->intsem,1,B_TIMEOUT,timeout);

	/* ---
		The Maxtor 7420AV (and maybe others?) seem to have some sort of
		watchdog timer, and if the host does not start doing something to
		the drive within 4-5 millisecs of an interrupt, it interrupts AGAIN!
		The interrupt_gate flag prevents getting more than one interrupt for
		the same 'event'.
	--- */

	if(err == B_NO_ERROR) {
		bi->interrupt_gate = 1;
		if(debug_interrupts)
			dprintf("IDE hpt366 -- intwait: got interrupt\n");
	}
	else {
		if(debug_interrupts)
			dprintf("IDE hpt366 -- intwait: error = %d (%s)\n", err, strerror(err));
	}
	return err;
}

static int32
inthand(bus_info *bus_info_p)
{
	uint16		sa = bus_info_p->busmaster_status_addr;
	uint8	bmstatus;
	if(debug_interrupts)
		dprintf("IDE hpt366 -- ide_inthand: interrupt for %x\n",
		        bus_info_p->command_block_base_addr);

	if(bus_info_p->shared_interrupt) {
		bmstatus = pci->read_io_8(sa);		// get busmaster status
		if(bmstatus & 4) {
			bus_info_p->busmasterstatus = bmstatus;	// save status
			pci->write_io_8(sa, bmstatus);	// clear interrupt and error bit
			if(debug_interrupts)
				dprintf("IDE hpt366 -- ide_inthand: got int for %x, intgate %d\n",
				        bus_info_p->command_block_base_addr,
				        bus_info_p->interrupt_gate);
		}
		else {
			if(debug_interrupts)
				dprintf("IDE hpt366 -- ide_inthand: "
				        "not for this bus %x (st = 0x%02x)\n",
				        bus_info_p->command_block_base_addr, bmstatus);
			return B_UNHANDLED_INTERRUPT;
		}
	}
	else {	// NonShared interrupts
		if(!bus_info_p->buslock) {
			if(debug_interrupts)
				dprintf("IDE hpt366 -- ide_inthand: not for this bus %x\n",
				        bus_info_p->command_block_base_addr);
			return B_UNHANDLED_INTERRUPT;
		}
		if(bus_info_p->dmainuse) {
			uint16 ba = bus_info_p->busmaster_base_addr;
			/* stop bus master */
			pci->write_io_8(ba+0x0, pci->read_io_8(ba+0x0) & 0xfe);
			/* get status */
			bus_info_p->busmasterstatus = pci->read_io_8(ba+2);
			/* clear interrupt and error bit */
			pci->write_io_8(ba+2, bus_info_p->busmasterstatus);
		}
	}
	/* read STATUS to clear interrupt */
	pci->read_io_8(bus_info_p->command_block_base_addr+7);

	if (bus_info_p->interrupt_gate) {		/* see comment in intwait() */
		bus_info_p->interrupt_gate = 0;
		release_sem_etc (bus_info_p->intsem,1,B_DO_NOT_RESCHEDULE);
	}
	else {
		// We have already handled the interrupt for this event.
		if(debug_interrupts)
			dprintf("IDE hpt366 -- ide_inthand: "
			        "interrupts issued multiple times for same event %x (st = 0x%02x)\n",
			        bus_info_p->command_block_base_addr, bmstatus);
		return B_UNHANDLED_INTERRUPT;
	}
	if(debug_interrupts)
		dprintf("IDE hpt366 -- ide_inthand: interrupt handled\n");
	return B_INVOKE_SCHEDULER;
}

static size_t
update_prd_table(bus_info *bi, const iovec *vec, size_t veccount,
                 uint32 startbyte, uint32 blocksize, size_t size)
{
	const int 		max_table = 32;
	const uint32	max_size = 0x10000;	/* max size per PRD */
	const uint32	boundary = 0x10000;	/* regions cannot cross this boundary */

	physical_entry	p[max_table];		/* table of physical blocks */
	int				taken = 0;			/* # bytes set up in dma list so far */
	int				i;					/* loop counter */

	prd_t			*prd = bi->prd_table;
	prd_t			*prdend = bi->prd_table+(prd_table_size/sizeof(prd_t))-1; 

//dprintf("IDE hpt366 -- update_prd_table: vec 0x%08x, %d, startbyte %d, size %d\n",
//        vec, veccount, startbyte, size);

	for (;;) {
		while(veccount > 0 && vec->iov_len <= startbyte) {
			startbyte -= vec->iov_len;
			vec++;
			veccount--;
		}
		if(veccount == 0) {
			if(debug_errors)
				dprintf("IDE hpt366 -- update_prd_table: should not be here\n");
			return 0;
		}
		if((((uint32)vec->iov_base+startbyte) & bi->bad_alignment_mask) ||
		   ((vec->iov_len-startbyte) & bi->bad_alignment_mask)) {
			if(debug_errors)
				dprintf("IDE hpt366 -- update_prd_table: bad alignment\n");
			return 0;
		}
		get_memory_map((uint8*)vec->iov_base+startbyte,
		               min(vec->iov_len-startbyte, size), p, max_table);
		for (i = 0; i < max_table; i++) {			/* for each phys block */
			int	psize;		/* size current physical block */
			
			psize = p[i].size;
			startbyte += psize;
			if(psize == 0)
				break;
			while (psize > 0) {							/* for each PRD */
				uint32	paddress;
				int count;		/* # usable bytes from current block */

				paddress = (uint32)pci->ram_address(p[i].address);
				count = min (psize, max_size);
				
				/* now check we don't cross any restricted boundary */
				if ((paddress & ~(boundary-1)) !=
				    ((paddress+count-1) & ~(boundary-1)))
					 count = boundary - (paddress & (boundary-1));
				
				prd->address = B_HOST_TO_LENDIAN_INT32(paddress);
				/* map 64k xfer to 0 transfer_count */
				prd->byte_count = B_HOST_TO_LENDIAN_INT16(count & 0xfffe);
				
//				dprintf("IDE hpt366 -- update_prd_table: "
//				        "added addr 0x%08x, byte count 0x%04x\n",
//				        prd->address, prd->byte_count);

				taken += count;
				size -= count;

				if ( (size == 0) || (prd == prdend) ) {
					int strip = taken % blocksize;
					while(strip > 0) {
						if(prd->byte_count > strip) {
							if(debug_prd)
								dprintf("IDE hpt366 -- update_prd_table: "
								        "resized addr 0x%08x, byte count "
								        "0x%04x to 0x%04x\n",
								        prd->address, prd->byte_count,
										prd->byte_count-strip);
							prd->byte_count -= strip;
							taken -= strip;
							strip = 0;
						}
						else {
							if(debug_prd)
								dprintf("IDE hpt366 -- update_prd_table: "
								        "removed addr 0x%08x, byte count 0x%04x\n",
								        prd->address, prd->byte_count);
							strip -= prd->byte_count;
							taken -= prd->byte_count;
							prd--;
						}
					}
					prd->eot = 1;
					return taken;
				}
				prd->eot = 0;

				prd++;
				p[i].address = (char *) p[i].address + count;
				psize -= count;
			}
		}
	}
	if(debug_prd)
		dprintf("IDE hpt366 -- update_prd_table returns gracefully\n");
}

static status_t
prepare_dma(bus_info *bi, const iovec *vec, size_t veccount, uint32 startbyte,
            uint32 blocksize, size_t *numBytes, bool to_device)
{
	uint16		ba = bi->busmaster_base_addr;
	uint32		reserved;
	size_t		newsize;

	if(disable_dma)
		return B_NOT_ALLOWED;
	if(debug_dma)
		dprintf("IDE hpt366 -- prepare_dma: 0x%08x, size=%d\n", startbyte, *numBytes);
	if(!bi->buslock) {
		if(debug_errors)
			dprintf("IDE hpt366 -- prepare_dma: bus not locked\n");
		return B_ERROR;
	}
	if(ba == 0) {
		// not dma capable
		return B_NOT_ALLOWED;
	}
	
	if(bi->dmainuse) {
		if(debug_errors)
			dprintf("IDE hpt366 -- prepare_dma: already active\n");
		return B_ERROR;
	}
	bi->dmainuse = true;
	bi->dma_started = false;
	bi->busmasterstatus = 0;

	/* how many bytes can the next dma table really give us? */
	newsize = update_prd_table(bi, vec, veccount, startbyte,
	                           blocksize, *numBytes);
	if(*numBytes != newsize) {
		if(newsize < *numBytes && newsize > 0) {
			if(debug_errors)
					dprintf("IDE hpt366 -- prepare_dma: Warning, PRD table exhausted\n");
		}
		else {
			if(debug_errors)
				dprintf("IDE hpt366 -- prepare_dma: Error, PRD table wrong\n");
			return B_ERROR;
		}
	}
	*numBytes = newsize;

	/* load the table address */
	reserved = B_LENDIAN_TO_HOST_INT32(pci->read_io_32(ba+0x04)) & 0x03;
	pci->write_io_32(ba+0x04,
	                 B_HOST_TO_LENDIAN_INT32(
	                    (uint32)pci->ram_address((void*)bi->prd_table_physical)
	                  | reserved));
	if(to_device) {
		/* set the bus master read mode  */
		pci->write_io_8(ba+0x0, pci->read_io_8(ba+0x0) & 0xf7);
	}
	else {
		/* set the bus master write mode  */
		pci->write_io_8(ba+0x0, pci->read_io_8(ba+0x0) | 0x8);
	}
	if(debug_dma)
		dprintf("IDE hpt366 -- prepare_dma returns gracefully\n");
	return B_NO_ERROR;
}

static status_t
start_dma(bus_info *bi)
{
	uint16 ba;
	if(disable_dma)
		return B_NOT_ALLOWED;
	if(debug_dma)
		dprintf("IDE hpt366 -- start dma\n");

	if(!bi->dmainuse) {
			if(debug_errors)
			dprintf("IDE hpt366 -- start_dma: start what???\n");
		return B_ERROR;
	}
	ba = bi->busmaster_base_addr;
	bi->busmasterstatus = 1;
	bi->dma_started = true;
	/* launch the dma on the pci side */
	pci->write_io_8(ba+0x0, pci->read_io_8(ba+0x0) | 0x01);
	if(debug_dma)
		dprintf("IDE hpt366 -- start_dma returns gracefully\n");
	return B_NO_ERROR;
}

static status_t
finish_dma(bus_info *bi)
{
	uint16 ba;
	if(disable_dma)
		return B_NOT_ALLOWED;
	if(debug_dma)
		dprintf("IDE hpt366 -- finish_dma\n");
	ba = bi->busmaster_base_addr;
	if(!bi->buslock) {
		if(debug_errors)
			dprintf("IDE hpt366 -- finish_dma: bus not locked\n");
		return B_ERROR;
	}
	if(ba == 0) {
		return B_NOT_ALLOWED;
	}
	if(!bi->dmainuse) {
		if(debug_errors)
			dprintf("IDE hpt366 -- finish_dma: finish what???\n");
		return B_ERROR;
	}
	bi->dmainuse = false;
	if(!bi->dma_started) {
		if(debug_errors)
			dprintf("IDE hpt366 -- finish_dma: dma was never started\n");
		return B_ERROR;
	}

	/*turn off the start/stop bit in the cmd reg. */
	pci->write_io_8(ba+0x0, pci->read_io_8(ba+0x0) & 0xfe);
	bi->dma_started = false;
	if(bi->busmasterstatus & 2)
		if(debug_errors)
			dprintf("IDE hpt366 -- finish_dma: memory transfer error\n");
		
	switch(bi->busmasterstatus & 0x5) {
		case 0:
			if(debug_errors)
				dprintf("IDE hpt366 -- finish_dma: DMA transfer failed\n");
			return B_ERROR;
		case 1:
			if(debug_errors)
				dprintf("IDE hpt366 -- finish_dma: DMA transfer aborted\n");
			return B_ERROR;
		case 4:
			return B_NO_ERROR;
	    case 5:
			// XXX what to do here?
			if(debug_errors)
				dprintf("IDE hpt366 -- finish_dma: PRD size > device transfer size!\n");
			return B_NO_ERROR;
			return B_INTERRUPTED;
	}
	return B_NO_ERROR;
}

static uint32
get_bad_alignment_mask(bus_info *bi)
{
	if(debug_verbose)
		dprintf("IDE hpt366 -- get_bad_alignment_mask returns %d\n",bi->bad_alignment_mask);
	return bi->bad_alignment_mask;
}

static status_t
get_dma_mode(bus_info *bi, bool device1, uint32 *mode)
{
	uint8 busclock;
	uint32 TimingRegister;
	uint8 regtime;
	uint8 linuxMode;
	if(disable_dma)
		return B_NOT_ALLOWED;
	if(debug_dma)
		dprintf("IDE hpt366 -- get_dma_mode\n");

	regtime		=  device1 ? 0x44 : 0x40; // 0=Master, >0=Slave

	busclock = 
		pci->read_pci_config(bi->pci_dev.bus, bi->pci_dev.device,
		                     bi->pci_dev.function, regtime+1,
		                     1);
	TimingRegister =
		pci->read_pci_config(bi->pci_dev.bus, bi->pci_dev.device,
		                     bi->pci_dev.function, regtime,
		                     4);

	switch(busclock) { // PCI bus clock
		case 0xd9: // 40MHz
			linuxMode = pci_dma_mode_list (TimingRegister, forty_base) & 0x000000FF;
			break;
		case 0x85: // 25MHz
			linuxMode = pci_dma_mode_list (TimingRegister, twenty_five_base) & 0x000000FF;
			break;
		case 0xa7: // 33MHz
		default:   // other
			linuxMode = pci_dma_mode_list (TimingRegister, thirty_three_base) & 0x000000FF;
			break;
	}
	if(linuxMode==0xFF)
	{
		dprintf("IDE hpt366 -- get_dma_mode Couldn't determine DMA mode - what to do?\n");
		return B_NOT_ALLOWED;
	}

	switch(linuxMode) {
		case XFER_UDMA_4:
			*mode = 0x14 & 0x000000FF; //udma mode 4 (udma66)
			break;
		case XFER_UDMA_3:
			*mode = 0x13 & 0x000000FF; //udma mode 3 (udma33)
			break;
		case XFER_UDMA_2:
			*mode = 0x12 & 0x000000FF; //udma mode 2
			break;
		case XFER_UDMA_1:
			*mode = 0x11 & 0x000000FF; //udma mode 1
			break;
		case XFER_UDMA_0:
			*mode = 0x10 & 0x000000FF; //udma mode 0
			break;
		case XFER_MW_DMA_2:
			*mode = 0x02 & 0x000000FF; //dma mode 2
			break;
		case XFER_MW_DMA_1:
			*mode = 0x01 & 0x000000FF; //dma mode 1
			break;
		case XFER_MW_DMA_0:
			*mode = 0x00 & 0x000000FF; //dma mode 0
			break;
		case XFER_PIO_4:
		case XFER_PIO_3:
		case XFER_PIO_2:
		case XFER_PIO_1:
		case XFER_PIO_0:
		case XFER_PIO_SLOW:
		default:
			*mode=0;
			if(debug_dma)
				dprintf("IDE hpt366 -- get_dma_mode - Was not in DMA mode - what to do?\n");
			return B_NOT_ALLOWED;
	}
	if(debug_dma)
		dprintf("IDE hpt366 -- get_dma_mode returns 0x%X\n",*mode);
	return B_NO_ERROR;
}

static status_t
set_dma_mode(bus_info *bi, bool device1, uint32 mode)
{
	uint8 linuxMode;
	if(disable_dma)
		return B_NOT_ALLOWED;
	if(debug_dma)
		dprintf("IDE hpt366 -- set_dma_mode (currently not enabled)\n");

	return B_NOT_ALLOWED;	// set_dma_mode is not officially supported yet, so we don't use it.
							// Uncomment this return if you want to test my version (provided Be's IDE module calls it).

	switch(mode) {
		case 0x00: // dma mode 0
			linuxMode = XFER_MW_DMA_0;
			break;
		case 0x01: // dma mode 1
			linuxMode = XFER_MW_DMA_1;
			break;
		case 0x02: // dma mode 2
			linuxMode = XFER_MW_DMA_2;
			break;
		case 0x10: // udma mode 0
			linuxMode = XFER_UDMA_0;
			break;
		case 0x11: // udma mode 1
			linuxMode = XFER_UDMA_0;
			break;
		case 0x12: // udma mode 2
			linuxMode = XFER_UDMA_0;
			break;
		case 0x13: // udma mode 3 (udma66)
			linuxMode = XFER_UDMA_0;
			break;
		case 0x14: // udma mode 4 (udma66)
			linuxMode = XFER_UDMA_0;
			break;
		default:
			dprintf("IDE hpt366 -- set_dma_mode - unsupported DMA mode\n");
			return B_NOT_ALLOWED;
	}
	hpt366_tune_chipset (bi, device1, linuxMode);
	return B_NO_ERROR;
}

/* -----
   create_prd_table_area creates and locks an area that will be used to store
   physical region descriptor tables. This area must be locked in physical
   memory to be accessed by dma controllers
----- */

static status_t
create_prd_table_area (bus_info *bi)
{
	status_t		err;
	physical_entry	prdt_physical_entry;

	if(debug_prd)
		dprintf("IDE hpt366 -- create_prd_table_area\n");
	bi->dmainuse = false;
	/* the area doesn't exist, try to create it */
	err = bi->prd_table_area =
		create_area( "PCI_IDE_HPT366_PRD_TABLE_AREA", (void**)&bi->prd_table,
		             B_ANY_KERNEL_ADDRESS,
		             prd_table_size, B_FULL_LOCK, B_READ_AREA | B_WRITE_AREA);
	if(err < B_NO_ERROR) {
		if(debug_errors)
			dprintf("IDE hpt366 -- create_prd_table_area: "
		    	    "couldn't create dma table area\n");
		goto err0;
	}
	memset(bi->prd_table, 0, prd_table_size); // clear reserved bits

	err = lock_memory(bi->prd_table, prd_table_size, B_DMA_IO | B_READ_DEVICE);
	if(err != B_NO_ERROR) {
		if(debug_errors)
			dprintf("IDE hpt366 -- create_prd_table_area: "
			        "dma table couldn't be locked.\n");
		goto err1;
	}

	err = get_memory_map(bi->prd_table, prd_table_size, &prdt_physical_entry,1);
	if(err != B_NO_ERROR) {
		if(debug_errors)
		dprintf("IDE hpt366 -- create_prd_table_area: "
		        "dma table area created & locked,\n"
		        "couldn't get memory map\n");
		goto err2;
	}
	bi->prd_table_physical = (uint32)prdt_physical_entry.address;

	return B_NO_ERROR;

err2:
	unlock_memory(bi->prd_table, prd_table_size, B_DMA_IO | B_READ_DEVICE);
err1:
	delete_area(bi->prd_table_area);
err0:
	if(debug_prd)
		dprintf("IDE hpt366 -- create_prd_table_area returns %d (%s)\n",err,strerror(err));
	return err;
}

/* -----
   delete_prd_table_area deletes a PRD table area
----- */

static int
delete_prd_table_area (bus_info *bi)
{
	int err;
	if(debug_prd)
		dprintf("IDE hpt366 -- delete_prd_table_area");
	if(bi->prd_table_area < 0)
	{
		if(debug_errors)
			dprintf("IDE hpt366 -- delete_prd_table_area returns error\n");
		return B_ERROR;
	}
	unlock_memory(bi->prd_table, prd_table_size, B_DMA_IO | B_READ_DEVICE);
	err=delete_area(bi->prd_table_area);
	if(debug_prd)
		dprintf("IDE hpt366 -- delete_prd_table_area returns %d\n",err);
	return err;
}

static status_t
find_devices()
{
	int			i;
	pci_info	h;
	status_t	err;
	bool		primary_controller_found = false;
	bool		secondary_controller_found = false;
	uint16 bm_base_reg;
	for(i=0; pci->get_nth_pci_info(i,&h) == B_NO_ERROR; i++ ) {
		uint32 full_id = h.vendor_id << 16 | h.device_id;
		bool is_hpt366;
		int controller_bus_count = 2;
		
		switch(full_id) {
			case 0x11030003: /* HighPoint HPT366 */
			case 0x11030004: /* HighPoint HPT366 */
				is_hpt366 = true;
				controller_bus_count = 1;
				break;
			default:
				is_hpt366 = false;
		}
		
		if(is_hpt366) {
			uint32 dma_bad_alignment_mask = 0x1;
			bus_info *newbus, *tmpbus;
			int i;
			newbus = NULL;
			newbus = malloc(sizeof(bus_info));
			if(newbus == NULL) {
				err = B_NO_MEMORY;
				goto err1;
			}
			newbus->pci_dev.bus = h.bus;
			newbus->pci_dev.device = h.device;
			newbus->pci_dev.function = h.function;

			dprintf("IDE hpt366 -- find_devices: found HighPoint HPT366 controller, 
			        Vendor=0x%04x, Device=0x%04x, Revision=0x%02x\n",
			        h.vendor_id, h.device_id, h.revision);

			h.class_api = 0x85;

			newbus->command_block_base_addr = h.u.h0.base_registers[0];
			newbus->control_block_register_addr = h.u.h0.base_registers[1] + 2;
			newbus->intnum = h.u.h0.interrupt_line;
			newbus->shared_interrupt = true;

			dprintf("IDE hpt366 -- find_devices: controller supports DMA\n");
			bm_base_reg = h.u.h0.base_registers[4];
			newbus->busmaster_base_addr = bm_base_reg;
			newbus->busmaster_status_addr = bm_base_reg + 2;
			newbus->bad_alignment_mask = dma_bad_alignment_mask;
			if(create_prd_table_area(newbus) != B_NO_ERROR) {
				dprintf("IDE hpt366 -- find_devices: disabled dma (couldn't create PRD table area)\n");
				newbus->prd_table_area = -1;
				newbus->busmaster_base_addr = 0;
			}
				
			newbus->mutex = -1;
			newbus->mutex_shared = false;
			newbus->buslock = false;
			newbus->intsem = -1;
			err = newbus->mutex = create_sem(1, "pciidebusmutex");
			if(err < B_NO_ERROR)
				goto err2;
			err = newbus->intsem = create_sem(0, "pciideintsem");
			if(err < B_NO_ERROR)
				goto err2;

			/* clear any pending interrupts */
			newbus->buslock = true; // no need to truly lock the bus here
			write_device_control(newbus, 0x0a); // disable interrupts
			pci->read_io_8(newbus->command_block_base_addr+7);	// read STATUS to clear interrupt
			newbus->buslock = false;

			newbus->interrupt_gate = 1;
			newbus->buslock = true; // no need to truly lock the bus here
			newbus->handler = (interrupt_handler)inthand;
			err = install_io_interrupt_handler(newbus->intnum,
												newbus->handler,
												newbus, 0);
			if(err != B_NO_ERROR) {
				dprintf("IDE hpt366 -- find_devices: "
				        "could not install interrupt handler (%d)\n",
				        newbus->intnum);
				delete_sem(newbus->mutex);
				delete_sem(newbus->intsem);
				free(newbus);
				dprintf("IDE hpt366 -- find_devices: skipped bus\n");
				continue;	// Next PCI device, please
			}
			write_device_control(newbus, 0x08); // enable interrupts again
			newbus->buslock = false;

			// add bus
			newbus->next=0;
			if(busses==0)
			{
				busses = newbus;
			}
			else
			{
				tmpbus=busses;
				while(tmpbus->next!=0)
					tmpbus=tmpbus->next;
				tmpbus->next=newbus;
			}

			buscount++;
			dprintf("IDE hpt366 -- find_devices: added bus,\n"
					"                         cmdblk: 0x%04x\n"
					"                         ctlreg: 0x%04x\n",
       				newbus->command_block_base_addr,
       				newbus->control_block_register_addr);
			continue;	// Next PCI device, please

err2:
			if(newbus->mutex >= 0 && !newbus->mutex_shared)
				delete_sem(newbus->mutex);
			if(newbus->intsem >= 0)
				delete_sem(newbus->intsem);
			delete_prd_table_area(newbus);
err1:
			if(newbus)
				free(newbus);
			return err;
		}	// if(is_hpt366)
	}	// Forall PCI devices
	return B_NO_ERROR;
}

hpt366_tune_all_busses(uint8 speed)
{
	if(busses)
	{
		hpt366_tune_chipset(busses, 0, speed);
		hpt366_tune_chipset(busses, 1, speed);
	}
	if(busses->next)
	{
		hpt366_tune_chipset(busses->next, 0, speed);
		hpt366_tune_chipset(busses->next, 1, speed);
	}	
}


static status_t
init()
{
	status_t err;
	static status_t uninit();
	void *settingsHandle;
	
	settingsHandle=load_driver_settings("hpt366");
	if(!settingsHandle)	
	{
		if(debug_errors)
			dprintf("IDE hpt366 -- init: Couldn't load settings - using default.\n");
	}
	else
	{
		debug_errors=get_driver_boolean_parameter(settingsHandle,"debug_errors",1,1);
		debug_settings=get_driver_boolean_parameter(settingsHandle,"debug_settings",1,1);
		debug_finddevices=get_driver_boolean_parameter(settingsHandle,"debug_finddevices",1,1);
		debug_interrupts=get_driver_boolean_parameter(settingsHandle,"debug_interrupts",0,1);
		debug_dma=get_driver_boolean_parameter(settingsHandle,"debug_dma",0,1);
		debug_prd=get_driver_boolean_parameter(settingsHandle,"debug_prd",0,1);
		debug_verbose=get_driver_boolean_parameter(settingsHandle,"debug_verbose",0,1);
		disable_dma=get_driver_boolean_parameter(settingsHandle,"disable_dma",0,1);
		strncpy(force_dma_mode,get_driver_parameter(settingsHandle,"force_dma_mode","AUTO","AUTO"),DMA_MODE_LENGTH);
		unload_driver_settings(settingsHandle);
		if(debug_settings)
		{
			dprintf("IDE hpt366 -- Setting debug_errors = %d\n",debug_errors);
			dprintf("IDE hpt366 -- Setting debug_settings = %d\n",debug_settings);
			dprintf("IDE hpt366 -- Setting debug_finddevices = %d\n",debug_finddevices);
			dprintf("IDE hpt366 -- Setting debug_interrupts = %d\n",debug_interrupts);
			dprintf("IDE hpt366 -- Setting debug_dma = %d\n",debug_dma);
			dprintf("IDE hpt366 -- Setting debug_prd = %d\n",debug_prd);
			dprintf("IDE hpt366 -- Setting debug_verbose = %d\n",debug_verbose);
			dprintf("IDE hpt366 -- Setting disable_dma = %d\n",disable_dma);
			dprintf("IDE hpt366 -- Setting force_dma_mode = %s\n",force_dma_mode);
		}
	}

	if(debug_verbose)
		dprintf("IDE hpt366 -- init\n");
	buscount = 0;
	busses = NULL;

	if(debug_verbose)
		dprintf("IDE hpt366 -- init: Getting PCI module\n");
	err = get_module(B_PCI_MODULE_NAME, (module_info **)&pci);
	if (err)
	{
		if(debug_errors)
			dprintf("IDE hpt366 -- init: Couldn't get PCI module\n");
		return err;
	}
	if(debug_verbose)
		dprintf("IDE hpt366 -- init: Finding devices\n");
	err = find_devices();
	if (err)
		goto err;
	if(debug_verbose)
		dprintf("IDE hpt366 -- init: Loading settings\n");

	if(strcmp(force_dma_mode,"AUTO"))
	{
		if(!strcmp(force_dma_mode,"DMA0"))
			hpt366_tune_all_busses(XFER_MW_DMA_0);
		else if(!strcmp(force_dma_mode,"DMA1"))
			hpt366_tune_all_busses(XFER_MW_DMA_1);
		else if(!strcmp(force_dma_mode,"DMA2"))
			hpt366_tune_all_busses(XFER_MW_DMA_2);
		else if(!strcmp(force_dma_mode,"UDMA0"))
			hpt366_tune_all_busses(XFER_UDMA_0);
		else if(!strcmp(force_dma_mode,"UDMA1"))
			hpt366_tune_all_busses(XFER_UDMA_1);
		else if(!strcmp(force_dma_mode,"UDMA2"))
			hpt366_tune_all_busses(XFER_UDMA_2);
		else if(!strcmp(force_dma_mode,"UDMA3"))
			hpt366_tune_all_busses(XFER_UDMA_3);
		else if(!strcmp(force_dma_mode,"UDMA4"))
			hpt366_tune_all_busses(XFER_UDMA_4);
	}

	if(debug_verbose)
		dprintf("IDE hpt366 -- init: done\n");
	return B_NO_ERROR;

err:
	uninit();
	if(debug_errors)
		dprintf("IDE hpt366 -- init: error (%s)- exiting\n",strerror(err));
	return err;
}

static status_t
uninit()
{
	status_t err;
	if(debug_verbose)
		dprintf("IDE hpt366 -- uninit\n");
	if (pci)
		put_module(B_PCI_MODULE_NAME);
	while(busses != NULL) {
		bus_info *bus = busses;
		busses = bus->next;

		bus->buslock = true; // no need to lock the bus here
		write_device_control(bus, 0x0a); // disable interrupts
		bus->buslock = false;

		if(debug_verbose)
			dprintf("IDE hpt366 -- uninit: remove inthand %d\n", bus->intnum);
		if (bus->handler != NULL) {
			err = remove_io_interrupt_handler(bus->intnum,
		                                  bus->handler, bus);
			if(err != B_NO_ERROR) {
				if(debug_errors)
				dprintf("IDE hpt366 -- uninit: "
				        "remove inthand (int %d) failed for bus at 0x%x\n",
				        bus->intnum, bus->command_block_base_addr);
			}
		}
		if(bus->mutex >= 0 && !bus->mutex_shared)
			delete_sem(bus->mutex);
		if(bus->intsem >= 0)
			delete_sem(bus->intsem);
		if(bus->prd_table_area >= 0)
			delete_prd_table_area(bus);
		free(bus);
	}
	return B_NO_ERROR;
}

static status_t
std_ops(int32 op, ...)
{
	switch(op) {
	case B_MODULE_INIT:
		return init();
	case B_MODULE_UNINIT:
		return uninit();
	default:
		return -1;
	}
}

static status_t
rescan()
{
	return 0;
}

ide_bus_info hpt366_module = {
	{
		{
			"busses/ide/hpt366/v0.6",	// 0.6 is the IDE API version.
			0,
			&std_ops
		},
		&rescan
	},
	(uint32 (*) (uint32))						&get_nth_cookie,
	&get_bus_count,
	(int32 (*) (uint32))						&get_abs_bus_num,
	(status_t (*) (uint32))						&acquire_bus,
	(status_t (*) (uint32))						&release_bus,
	(status_t (*) (uint32,
	 ide_task_file*,ide_reg_mask mask))			&write_command_block_regs,
	(status_t (*)(uint32,
	 ide_task_file*,ide_reg_mask mask))			&read_command_block_regs,
	(uint8 (*) (uint32))						&get_altstatus,
	(void (*) (uint32,uint8))					&write_device_control,
	(void (*) (uint32,uint16*,uint16))			&write_pio_16,
	(void (*) (uint32,uint16*,uint16))			&read_pio_16,
	(status_t (*) (uint32,bigtime_t))			&intwait,
	(status_t (*)(uint32,const iovec *,size_t,
	              uint32,uint32,size_t*,bool))	&prepare_dma,
	(status_t (*)(uint32))						&start_dma,
	(status_t (*)(uint32))						&finish_dma,
	(uint32 (*)(uint32))						&get_bad_alignment_mask,
	(status_t (*)(uint32,bool,uint32*))			&get_dma_mode,
	(status_t (*)(uint32,bool,uint32))			&set_dma_mode
};

_EXPORT 
ide_bus_info *modules[] = {
	&hpt366_module,
	NULL
};
